from itertools import combinations
from typing import Tuple
import numpy as np

from luma.interface.super import Estimator, Evaluator, Transformer, Supervised
from luma.interface.util import Matrix, Vector, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.model_selection.split import TrainTestSplit
from luma.model_selection.cv import CrossValidator


__all__ = (
    'SBS',
    'SFS',
    'RFE'
)


class SBS(Transformer, Transformer.Feature, Supervised):
    
    """
    Sequential Backward Selection (SBS) is a feature selection technique 
    used in machine learning. It starts with all features and iteratively
    removes the least significant feature at each step_size. The goal is to 
    reduce dimensionality while maintaining or improving model performance. 
    The process continues until the desired number of features is reached 
    or performance criteria are met.
    
    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `test_size` : Proportional size of the validation set
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `random_state` : Seed for splitting the data
    
    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * For `metric`, both class or instance are possible
    
    Examples
    --------
    >>> sbs = SBS(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  test_size=0.2,
                  cv=5,
                  random_state=None)
    >>> sbs.fit(X, y)
    >>> Z = sbs.transform(X)
    
    """
    
    def __init__(self, 
                 estimator: Estimator = None,
                 n_features: int | float = 1,
                 metric: Evaluator = None,
                 test_size: float = 0.2,
                 cv: int = 5,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.test_size = test_size
        self.cv = cv
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'SBS':
        _, n = X.shape
        self.estimator = Clone(self.estimator).get
        
        if 0 < self.n_features < 1:
            self.n_features = np.ceil(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)
        
        if self.cv:
            cv_model = CrossValidator(estimator=self.estimator,
                                      metric=self.metric, 
                                      cv=self.cv,
                                      random_state=self.random_state)
        else:
            Xy_split = TrainTestSplit.split(X, y, 
                                            test_size=self.test_size,
                                            random_state=self.random_state)
            X_train, X_test, y_train, y_test = Xy_split

        self.indices = tuple(range(n))
        self.subsets = [self.indices]
        
        if self.cv: score = cv_model.score(X, y)
        else: score = self._calculate_score(X_train, X_test, 
                                            y_train, y_test, indices=self.indices)
        self.scores = [score]
        
        iter = 1
        while n > self.n_features:
            scores = []
            subsets = []
            for p in combinations(self.indices, n - 1):
                if self.cv: score = cv_model.score(X[:, p], y)
                else: score = self._calculate_score(X_train, X_test, 
                                                    y_train, y_test, indices=p)
                scores.append(score)
                subsets.append(p)
            
            best = np.argmax(scores)
            self.indices = subsets[best]
            self.subsets.append(self.indices)
            self.scores.append(scores[best])
            
            if self.verbose:
                print(f'[SBS] Feature added: {self.indices[-1]}', 
                      f'with score: {scores[best]}')
            
            n -= 1
            iter += 1
        
        self.n_score = self.scores[-1]
        if self.verbose:
            print(f'[SBS] Selection finished with final features: {self.indices}')
        
        self._fitted = True
        return self
    
    def _calculate_score(self, 
                         X_train: Matrix, X_test: Matrix, 
                         y_train: Matrix, y_test: Matrix, indices: Tuple) -> float:
        self.estimator.fit(X_train[:, indices], y_train)
        y_pred = self.estimator.predict(X_test[:, indices])
        
        return self.metric.compute(y_true=y_test, y_pred=y_pred)
    
    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return X[:, self.indices]
    
    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)
    
    def set_params(self, 
                   estimator: Estimator = None,
                   n_features: int | float = None,
                   metric: Evaluator = None,
                   test_size: float = None,
                   cv: int = None,
                   random_state: int = None) -> None:
        if estimator is not None: self.estimator = estimator
        if metric is not None: self.metric = metric
        if test_size is not None: self.test_size = float(test_size)
        if cv is not None: self.cv = int(cv)
        if random_state is not None: self.random_state = int(random_state)
        if n_features is not None:
            if 0 < n_features < 1: self.n_features = float(n_features)
            else: self.n_features = int(n_features)


class SFS(Transformer, Transformer.Feature, Supervised):
    
    """
    Sequential Forward Selection (SFS) is a feature selection technique
    used in machine learning. It starts with no features and iteratively
    adds the most significant feature at each step. The goal is to
    reduce dimensionality while maintaining or improving model performance.
    The process continues until the desired number of features is reached
    or performance criteria are met.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `test_size` : Proportional size of the validation set
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `random_state` : Seed for splitting the data
    
    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * Do not use the estimator for `SFS` if it is supposed to be the
        main estimator for the model
    * For `metric`, both class or instance are possible
    
    Examples
    --------
    >>> sfs = SFS(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  test_size=0.2,
                  cv=5,
                  random_state=None)
    >>> sfs.fit(X, y)
    >>> Z = sbs.transform(X)
    
    """

    def __init__(self, 
                 estimator: Estimator = None,
                 n_features: int | float = 1,
                 metric: Evaluator = None,
                 test_size: float = 0.2,
                 cv: int = 5,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.test_size = test_size
        self.cv = cv
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> 'SFS':
        _, n = X.shape
        self.estimator = Clone(self.estimator).get
        
        if isinstance(self.n_features, float) and 0 < self.n_features < 1:
            self.n_features = np.ceil(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)

        if self.cv:
            cv_model = CrossValidator(estimator=self.estimator,
                                      metric=self.metric, 
                                      cv=self.cv,
                                      random_state=self.random_state)
        else:
            Xy_split = TrainTestSplit.split(X, y, 
                                            test_size=self.test_size,
                                            random_state=self.random_state)
            X_train, X_test, y_train, y_test = Xy_split

        self.indices = ()
        self.subsets = [self.indices]
        score = -np.inf
        self.scores = [score]

        while len(self.indices) < self.n_features:
            scores = []
            subsets = []
            for p in combinations(range(n), len(self.indices) + 1):
                if set(self.indices).issubset(set(p)):
                    if self.cv: score = cv_model.score(X[:, p], y)
                    else: score = self._calculate_score(X_train, X_test, 
                                                        y_train, y_test, indices=p)
                    scores.append(score)
                    subsets.append(p)

            best = np.argmax(scores)
            self.indices = subsets[best]
            self.subsets.append(self.indices)
            self.scores.append(scores[best])

            if self.verbose:
                print(f'[SFS] Feature added: {self.indices[-1]}', 
                      f'with score: {scores[best]}')

        self.n_score = self.scores[-1]
        if self.verbose:
            print(f'[SFS] Selection finished with final features: {self.indices}')

        self._fitted = True
        return self
    
    def _calculate_score(self, 
                         X_train: Matrix, X_test: Matrix, 
                         y_train: Matrix, y_test: Matrix, indices: Tuple) -> float:
        self.estimator.fit(X_train[:, indices], y_train)
        y_pred = self.estimator.predict(X_test[:, indices])
        
        return self.metric.compute(y_true=y_test, y_pred=y_pred)
    
    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return X[:, self.indices]
    
    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)
    
    def set_params(self, 
                   estimator: Estimator = None,
                   n_features: int | float = None,
                   metric: Evaluator = None,
                   test_size: float = None,
                   cv: int = None,
                   random_state: int = None) -> None:
        if estimator is not None: self.estimator = estimator
        if metric is not None: self.metric = metric
        if test_size is not None: self.test_size = float(test_size)
        if cv is not None: self.cv = int(cv)
        if random_state is not None: self.random_state = int(random_state)
        if n_features is not None:
            if 0 < n_features < 1: self.n_features = float(n_features)
            else: self.n_features = int(n_features)


class RFE(Transformer, Transformer.Feature, Supervised):
    
    """
    Recursive Feature Elimination (RFE) is a feature selection method 
    used in machine learning. It works by recursively removing the least 
    important features based on a given model's feature importance or 
    coefficients. RFE starts with all features and eliminates a specified 
    number or fraction of features at each step. The process continues 
    until the desired number of features is reached. This method helps in 
    enhancing model performance and interpretability by reducing complexity.
    
    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `step_size` : Number of features to eliminate in each step
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `random_state` : Seed for splitting the data
    
    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * Do not use the estimator for `RFE` if it is supposed to be the
        main estimator for the model
    * For `metric`, both class or instance are possible
    
    Examples
    --------
    >>> rfe = RFE(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  step_size=1,
                  cv=5,
                  random_state=None)
    >>> rfe.fit(X, y)
    >>> Z = rfe.transform(X)
    
    """
    
    def __init__(self,
                 estimator: Estimator = None,
                 n_features: int | float = 1,
                 metric: Evaluator = None,
                 step_size: int = 1,
                 cv: int = 5,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.step_size = step_size
        self.cv = cv
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> 'RFE':
        _, n = X.shape
        self.estimator = Clone(self.estimator).get
        
        if 0 < self.n_features < 1:
            self.n_features = np.floor(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)
        
        self.cv_model = CrossValidator(estimator=self.estimator, 
                                       metric=self.metric,
                                       cv=self.cv,
                                       random_state=self.random_state)

        self.support = np.ones(n, dtype=bool)
        self.ranking = np.ones(n, dtype=int)

        while np.sum(self.support) > self.n_features:
            scores = self._calculate_score(X, y)
            indices = np.argsort(scores)[::-1][:self.step_size]
            self.support[indices] = False
            self.ranking[indices] += 1

            if self.verbose:
                print(f'[RFE] Removed {self.step_size} features,', 
                      f'remaining features: {np.sum(self.support)}')
        
        features = tuple(i for i, sup in enumerate(self.support) if sup)
        if self.verbose:
            print(f'[RFE] Elimination finished with final features: {features}')
        
        self._fitted = True
        return self

    def _calculate_score(self, X: Matrix, y: Vector) -> Vector[float]:
        scores = np.zeros(X.shape[1])
        for i in range(X.shape[1]):
            if not self.support[i]: continue
            indices = self.support.copy()
            indices[i] = False
            
            if self.cv:
                score = self.cv_model.score(X[:, indices], y)
            else:
                self.estimator.fit(X[:, indices], y)
                y_pred = self.estimator.predict(X[:, indices])
                score = self.metric.compute(y_true=y, y_pred=y_pred)
                
            scores[i] = score
        
        return scores

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return X[:, self.support]

    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)
    
    def set_params(self, 
                   estimator: Estimator = None,
                   n_features: int | float = None,
                   metric: Evaluator = None,
                   step_size: int = None,
                   cv: int = None,
                   random_state: int = None) -> None:
        if estimator is not None: self.estimator = estimator
        if metric is not None: self.metric = metric
        if step_size is not None: self.test_size = int(step_size)
        if cv is not None: self.cv = int(cv)
        if random_state is not None: self.random_state = int(random_state)
        if n_features is not None:
            if 0 < n_features < 1: self.n_features = float(n_features)
            else: self.n_features = int(n_features)




from typing import Any, Callable, Literal

from scipy.spatial.distance import pdist, cdist, squareform
from scipy.sparse.csgraph import shortest_path
from scipy.spatial import KDTree
from scipy.linalg import eigh, svd
import numpy as np

from luma.interface.util import Matrix
from luma.reduction.linear import PCA
from luma.interface.super import Transformer, Unsupervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.clustering.kmeans import *
from luma.metric.distance import *


__all__ = (
    'TSNE', 
    'MDS', 
    'MetricMDS', 
    'LandmarkMDS', 
    'LLE',
    'ModifiedLLE', 
    'HessianLLE', 
    'SammonMapping', 
    'LaplacianEigenmap',
    'Isomap', 
    'ConformalIsomap', 
    'LTSA'
)


class TSNE(Transformer, Unsupervised):
    
    """
    t-SNE, which stands for t-distributed Stochastic Neighbor Embedding, 
    is a machine learning technique commonly used for data visualization 
    and dimensionality reduction.
    
    Parameters
    ----------
    `n_components` : Dimensionality of lower spcae
    `max_iter` : Number of iteration
    `learning_rate` : Updating factor of embedding optimization
    `perplexity` : Perplexity parameter of Gaussian kernel
    `verbose` : Provided details of each iteration when set `True`
    
    """
    
    def __init__(self, 
                 n_components: int = None, 
                 max_iter: int = 1000, 
                 learning_rate: int = 300, 
                 perplexity: int = 30,
                 verbose: bool = False) -> None:
        self.n_components = n_components
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.perplexity = perplexity
        self.verbose = verbose
        self._fitted = False
        
    def fit(self, X: Matrix) -> 'TSNE':
        size = X.shape[0]
        P = self._P_joint_probabilities(X)
        y = np.random.normal(0.0, 1e-4, (size, self.n_components))
        Y = [y, y]

        for i in range(1, self.max_iter + 1):
            Q = self._Q_joint_probabilities(Y[-1])
            gradient = self._compute_gradient(P, Q, Y[-1])
            y = Y[-1] - self.learning_rate * gradient
            y += self._momentum(i) * (Y[-1] - Y[-2])
            Y.append(y)
            
            if i % 10 == 0: Q = np.maximum(Q, 1e-12)
            if self.verbose and i % 50 == 0: 
                print(f'[t-SNE] main iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {np.linalg.norm(gradient)}')
            
        self.y = y
        self._fitted = True
        return self
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.y
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def _P_conditional(self, distances: Matrix, sigmas: Matrix) -> Matrix:
        exponent = np.exp(-distances / (2 * np.square(sigmas.reshape((-1,1)))))
        np.fill_diagonal(exponent, 0.0)
        exponent += 1e-8
        
        return exponent / exponent.sum(axis=1).reshape([-1,1])
    
    def _P_joint_probabilities(self, X: Matrix) -> Matrix:
        size = X.shape[0]
        distances = self._compute_pairwise_dist(X)
        sigmas = self._find_sigma_vector(distances)
        conditional_prob = self._P_conditional(distances, sigmas)
        
        return (conditional_prob + conditional_prob.T) / (2.0 * size)
    
    def _Q_joint_probabilities(self, y: Matrix) -> Matrix:
        distances = self._compute_pairwise_dist(y)
        nominator = 1 / (1 + distances)
        np.fill_diagonal(nominator, 0.0)
        
        return nominator / np.sum(np.sum(nominator))
    
    def _find_sigma_vector(self, distances: Matrix) -> Matrix:
        N = distances.shape[0]
        opt_sigmas = np.zeros(N)
        
        for i in range(N):
            func = lambda sig: self._compute_perplexity(
                self._P_conditional(distances[i:i + 1, :], np.array([sig])))
            
            opt_sigmas[i] = self._binary_search(func)
            if self.verbose and i % 50 == 0: 
                print(f'[t-SNE] sigma iteration: {i}/{N}', end='')
                print(f' - sigma-norm: {np.linalg.norm(opt_sigmas[i])}')
            
        if self.verbose: print('-' * 70)
        return opt_sigmas
    
    def _binary_search(self, func: Callable, 
                       tol: float = 1e-10, max_iter: int = 1000, 
                       low: float = 1e-10, high: float = 1e3) -> Matrix:
        for _ in range(max_iter):
            guess = (high + low) / 2.0
            val = func(guess)
            if val > self.perplexity: high = guess
            else: low = guess
            if np.abs(val - self.perplexity) <= tol: 
                return guess
        return guess
    
    def _compute_pairwise_dist(self, X: Matrix) -> Matrix:
        return np.sum((X[None, :] - X[:, None]) ** 2, axis=2)
    
    def _compute_gradient(self, P: Matrix, Q: Matrix, y: Matrix) -> Matrix:
        pq_diff = P - Q
        y_diff = np.expand_dims(y, axis=1) - np.expand_dims(y, axis=0)
        distances = self._compute_pairwise_dist(y)
        aux = 1 / (1 + distances)
        
        return 4 * (np.expand_dims(pq_diff, 2) * y_diff * np.expand_dims(aux, 2)).sum(axis=1)
    
    def _compute_perplexity(self, cond_matrix: Matrix) -> Matrix:
        entropy = -np.sum(cond_matrix * np.log2(cond_matrix), axis=1)
        return 2 ** entropy
    
    def _momentum(self, iteration: int):
        return 0.5 if iteration < 250 else 0.8
    
    def set_params(self,
                   n_components: int = None,
                   perplexity: float = None,
                   learning_rate: float = None,
                   max_iter: int = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if perplexity is not None: self.perplexity = float(perplexity)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)


class MDS(Transformer, Unsupervised):
    
    """
    Multidimensional Scaling (MDS) is a data analysis technique that visualizes 
    high-dimensional data in a lower-dimensional space while preserving the 
    relationships between data points. It uses a distance matrix to find a 
    lower-dimensional representation, making it easier to understand and 
    explore data patterns.
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    
    """
    
    def __init__(self, n_components: int = None) -> None:
        self.n_components = n_components
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'MDS':
        D = self._compute_dissimilarity(X)
        n = D.shape[0]
        H = np.eye(n) - np.ones((n, n)) / n
        B = -0.5 * H.dot(D).dot(H)
        
        eigvals, eigvecs = np.linalg.eigh(B)
        sorted_indices = np.argsort(eigvals)[::-1]
        self.eigvals = eigvals[sorted_indices]
        self.eigvecs = eigvecs[:, sorted_indices]
        
        self.eigvals = self.eigvals[:self.n_components]
        self.eigvecs = self.eigvecs[:, :self.n_components]
        self._fitted = True
        self.stress = self._compute_stress(D)
        return self
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        sqrt_eigvals = np.diag(np.sqrt(np.maximum(self.eigvals, 0)))
        return self.eigvecs.dot(sqrt_eigvals)
    
    def fit_transform(self, X) -> Any:
        self.fit(X)
        return self.transform()
    
    def _compute_dissimilarity(self, X: Matrix) -> Matrix:
        return squareform(pdist(X, metric='euclidean'))

    def _compute_stress(self, D: Matrix) -> float:
        Z = self.transform()
        stress = np.sum((D - self._compute_dissimilarity(Z)) ** 2)
        stress /= np.sum(D ** 2)
        stress = np.sqrt(stress)
        return stress
    
    def set_params(self, n_components: int = None) -> None:
        if n_components is not None: self.n_components = int(n_components)


class MetricMDS(Transformer, Unsupervised):
    
    """
    Metric Multidimensional Scaling (MDS) is a data analysis method that focuses 
    on accurately preserving the original pairwise distances or dissimilarities
    between objects in a high-dimensional space when projecting them onto a 
    lower-dimensional space. 
    
    Parameter
    ---------
    `n_components` : Dimensionality of low-space
    `metric` : Distance metric 
    (e.g. `euclidean`, `manhattan`, `chebyshev`, `minkowski`,
    `cos_similarity`, `correation`, `mahalanobis`)
    `p` : Power factor for `minkowski`
    
    """
    
    def __init__(self,
                 n_components: int = None,
                 p: float | int = None,
                 metric: Literal['euclidean', 'manhattan', 'chebyshev',
                                 'minkowski', 'cos_similarity', 'correlation',
                                 'mahalanobis'] = 'euclidean') -> None:
        self.n_components = n_components
        self.metric = metric
        self.p = p
        self._fitted = False

    def fit(self, X: Matrix) -> 'MetricMDS':
        D = self._compute_dissimilarity(X)
        n = D.shape[0]
        H = np.eye(n) - np.ones((n, n)) / n
        B = -0.5 * H.dot(D).dot(H)
        
        eigvals, eigvecs = np.linalg.eigh(B)
        sorted_indices = np.argsort(eigvals)[::-1]
        self.eigvals = eigvals[sorted_indices]
        self.eigvecs = eigvecs[:, sorted_indices]
        
        self.eigvals = self.eigvals[:self.n_components]
        self.eigvecs = self.eigvecs[:, :self.n_components]
        self._fitted = True
        self.stress = self._compute_stress(D)
        return self
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        sqrt_eigvals = np.diag(np.sqrt(np.maximum(self.eigvals, 0)))
        return self.eigvecs.dot(sqrt_eigvals)
    
    def fit_transform(self, X) -> Any:
        self.fit(X)
        return self.transform()
    
    def _compute_dissimilarity(self, X: Matrix) -> Matrix:
        m, _ = X.shape
        cov = np.cov(X, rowvar=False)
        D = np.ones((m, m))
        for i in range(m):
            for j in range(i + 1, m):
                join = (X[i], X[j])
                if self.metric == 'euclidean': d = Euclidean.distance(*join)
                elif self.metric == 'manhattan': d = Manhattan.distance(*join)
                elif self.metric == 'chebyshev': d = Chebyshev.distance(*join)
                elif self.metric == 'minkowski': d = Minkowski.distance(*join, self.p)
                elif self.metric == 'cos_similarity': d = CosineSimilarity.distance(*join)
                elif self.metric == 'correlation': d = Correlation.distance(*join)
                elif self.metric == 'mahalanobis': d = Mahalanobis.distance(*join, cov)
                else: raise ValueError(f'[mMDS] Invalid distance metric: {self.metric}')
                D[i, j] = D[j, i] = d
        
        return D

    def _compute_stress(self, D: Matrix) -> float:
        Z = self.transform()
        stress = np.sum((D - self._compute_dissimilarity(Z)) ** 2)
        stress /= np.sum(D ** 2)
        stress = np.sqrt(stress)
        return stress
    
    def set_params(self, 
                   n_components: int = None,
                   p: float | int = None,
                   metric: Literal = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if p is not None: self.p = float(p)
        if metric is not None: self.metric = str(metric)


class LandmarkMDS(Transformer, Unsupervised):
    
    """
    Landmark Multi-Dimensional Scaling (LMDS) is a dimensionality reduction method 
    that selects a subset of "landmark" points to efficiently approximate the 
    lower-dimensional representation of a dataset while preserving pairwise distances. 
    It simplifies the computation by focusing on distances between data points and 
    landmarks, making it suitable for large datasets.
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    `n_landmarks` : Number of landmarks
    `method` : Algorithm for landmark initialization
    (e.g. `random`, `kmeans`, `kmeans++`, `kmedians`)
    
    """
    
    def __init__(self, 
                 n_components: int = None, 
                 n_landmarks: int = 10,
                 method: Literal['random', 'kmeans', 'kmeans++', 'kmedians'] = 'random',
                 verbose: bool = False) -> None:
        self.n_components = n_components
        self.n_landmarks = n_landmarks
        self.method = method
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'LandmarkMDS':
        self.landmarks = self._initialize_landmarks(X)
        D = self._compute_dissimilarity(self.landmarks)
        n = D.shape[0]
        H = np.eye(n) - np.ones((n, n)) / n
        B = -0.5 * H.dot(D).dot(H)
        
        eigvals, eigvecs = np.linalg.eigh(B)
        sorted_indices = np.argsort(eigvals)[::-1]
        self.eigvals = eigvals[sorted_indices]
        self.eigvecs = eigvecs[:, sorted_indices]
        
        self.eigvals = self.eigvals[self.n_components]
        self.eigvecs = self.eigvecs[:, :self.n_components]
        
        D_mean = np.mean(D, axis=0)
        D_xland = np.sum((X[:, np.newaxis, :] - self.landmarks) ** 2, axis=2)
        self.mean_diff = D_mean - D_xland
        self._fitted = True
        return self
    
    def _initialize_landmarks(self, X: Matrix) -> Matrix:
        m, _ = X.shape
        if self.method == 'random':
            if self.verbose: print(f'[LMDS] Initializing landmarks with random choice')
            landmark = X[np.random.choice(m, self.n_landmarks, replace=False)]
            
        elif self.method == 'kmeans':
            if self.verbose: print(f'[LMDS] Initializing landmarks with K-Means centroids')
            kmeans = KMeansClustering(n_clusters=self.n_landmarks, verbose=self.verbose)
            kmeans.fit(X)
            landmark = kmeans.centroids
        
        elif self.method == 'kmeans++':
            if self.verbose: print(f'[LMDS] Initializing landmarks with K-Means++ centroids')
            kmeanspp = KMeansClusteringPlus(n_clusters=self.n_landmarks, verbose=self.verbose)
            kmeanspp.fit(X)
            landmark = kmeanspp.centroids
        
        elif self.method == 'kmedians':
            if self.verbose: print(f'[LMDS] Initializing landmarks with K-Medians medians')
            kmedians = KMediansClustering(n_clusters=self.n_landmarks, verbose=self.verbose)
            kmedians.fit(X)
            landmark = kmedians.medians
            
        else: raise UnsupportedParameterError(self.method)
        return landmark
    
    def _compute_dissimilarity(self, X: Matrix) -> Matrix:
        return squareform(pdist(X, metric='euclidean'))
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        L_sharp = self.eigvecs / np.sqrt(self.eigvals)
        return 0.5 * self.mean_diff.dot(L_sharp)
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def set_params(self,
                   n_components: int = None,
                   n_landmarks: int = None,
                   method: Literal = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if n_landmarks is not None: self.n_landmarks = int(n_landmarks)
        if method is not None: self.method = str(method)


class LLE(Transformer, Unsupervised):
    
    """
    Locally Linear Embedding (LLE) is a dimensionality reduction technique that aims 
    to find a lower-dimensional representation of data while preserving the local 
    relationships between data points. It works by approximating each data point 
    as a linear combination of its nearest neighbors, revealing the underlying 
    structure of the data in a lower-dimensional space. 
    
    Parameters
    ----------
    `n_neighbors` : Number of neighbors to be considered 'close'
    `n_components` : Dimensionality of low-space
    
    """
    
    def __init__(self, 
                 n_components: int = None, 
                 n_neighbors: int = 5, 
                 verbose: bool = False) -> None:
        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'LLE':
        m, _ = X.shape
        distances = np.zeros((m, m))
        for i in range(m):
            for j in range(m):
                distances[i, j] = np.linalg.norm(X[i] - X[j])

        W = np.zeros((m, m))
        neighbors = np.argsort(distances, axis=1)[:, 1:self.n_neighbors + 1]
        for i in range(m):
            Xi = X[neighbors[i]] - X[i]
            Z = np.dot(Xi, Xi.T)
            Z += np.pi * 1e-3 * np.identity(self.n_neighbors)
            
            w = np.linalg.solve(Z, np.ones(self.n_neighbors))
            w /= np.sum(w)
            W[i, neighbors[i]] = w
            
            if self.verbose and i % 100 == 0:
                print(f'[LLE] Optimized weight for instance {i}/{m}', end='')
                print(f' - weight-norm: {np.linalg.norm(w)}')
            
        M = np.identity(m) - W
        self.eigvals, self.eigvecs = np.linalg.eig(np.dot(M.T, M))
        self._fitted = True
        return self
        
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        indices = np.argsort(self.eigvals)[1:self.n_components + 1]
        return self.eigvecs[:, indices]
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()

    def set_params(self, 
                   n_neighbors: int = None, 
                   n_components: int = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)
        if n_components is not None: self.n_components = int(n_components)


class ModifiedLLE(Transformer, Unsupervised):
    
    """
    Modified Locally Linear Embedding (MLLE) is a dimensionality reduction technique 
    that extends the LLE method by introducing regularization for improved stability. 
    MLLE aims to find a lower-dimensional representation of data while preserving the 
    local relationships between data points. It works by approximating each data 
    point as a linear combination of its nearest neighbors, revealing the underlying 
    structure of the data in a lower-dimensional space. 

    Parameters
    ----------
    `n_neighbors` : Number of neighbors to be considered 'close'
    `n_components` : Dimensionality of the lower-dimensional space
    `regularization` : Regularization parameter for stability
    
    """

    def __init__(self, 
                 n_neighbors: int = 5, 
                 n_components: int = None, 
                 regularization: float = 1e-3, 
                 verbose: bool = False) -> None:
        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix) -> 'ModifiedLLE':
        m, _ = X.shape
        distances = np.zeros((m, m))
        for i in range(m):
            for j in range(m):
                distances[i, j] = np.linalg.norm(X[i] - X[j])

        W = np.zeros((m, m))
        neighbors = np.argsort(distances, axis=1)[:, 1 : self.n_neighbors + 1]
        for i in range(m):
            Xi = X[neighbors[i]] - X[i]
            Z = np.dot(Xi, Xi.T)
            Z += self.regularization * np.identity(self.n_neighbors)

            w = np.linalg.solve(Z, np.ones(self.n_neighbors))
            w /= np.sum(w)
            W[i, neighbors[i]] = w

            if self.verbose and i % 100 == 0:
                print(f'[ModifiedLLE] Optimized weight for instance {i}/{m}', end='')
                print(f' - weight-norm: {np.linalg.norm(w)}')

        M = np.identity(m) - W
        self.eigvals, self.eigvecs = np.linalg.eig(np.dot(M.T, M))
        self._fitted = True
        return self

    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        indices = np.argsort(self.eigvals)[1 : self.n_components + 1]
        return self.eigvecs[:, indices]

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def set_params(self, 
                   n_neighbors: int = None, 
                   n_components: int = None,
                   regularization: float = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)
        if n_components is not None: self.n_components = int(n_components)
        if regularization is not None: self.regularization = float(regularization)


class HessianLLE(Transformer, Unsupervised):
    
    """
    Hessian Locally Linear Embedding (Hessian LLE) is a dimensionality reduction 
    technique that extends Locally Linear Embedding (LLE). It aims to reduce the 
    dimensionality of high-dimensional data while preserving local linear relationships 
    and capturing the curvature of these relationships using the Hessian matrix.
    
    Parameters
    ----------
    `n_neighbors` : Number of neighbors to be considered 'close'
    `n_components` : Dimensionality of the lower-dimensional space
    `regularization` : Regularization parameter for stability
    
    """
    
    def __init__(self, 
                 n_neighbors: int = 5, 
                 n_components: int = None,
                 regularization: float = 1e-5,
                 verbose: bool = False) -> None:
        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'HessianLLE':
        m, _ = X.shape
        dp = self.n_components * (self.n_components + 1) // 2
        
        if self.n_neighbors <= self.n_components + dp:
            self.n_neighbors = self.n_components * (self.n_components + 3)
            self.n_neighbors /= 2
            print(f"[HessianLLE] n_neighbors set to {self.n_neighbors}",
                  "due to Hessian condition mismatch.")
        
        index_mat = self._construct_graph(X)
        W = np.zeros((dp * m, m))
        Yi = np.empty((self.n_neighbors, self.n_components + dp + 1))
        Yi[:, 0] = 1
        for i in range(m):
            neighbors = index_mat[i]
            Gi = X[neighbors]
            Gi -= Gi.mean(axis=0)
            
            U, _, _ = np.linalg.svd(Gi, full_matrices=0)
            Yi[:, 1:self.n_components + 1] = U[:, :self.n_components]
            j = self.n_components + 1
            for k in range(self.n_components):
                Yi[:, j:self.n_components + j - k] = \
                    U[:, k:k + 1] * U[:, k:self.n_components]
                j += self.n_components - k
            
            Q, R = np.linalg.qr(Yi)
            w = np.array(Q[:, self.n_components + 1:])
            S = w.sum(axis=0)
            S[np.where(np.abs(S) < self.regularization)] = 1.0
            w /= S
            
            xn, yn = np.meshgrid(neighbors, neighbors)
            W[xn, yn] += w.dot(w.T)
            
            if self.verbose and i % 100 == 0:
                print(f'[HessianLLE] Optimized weight for instance {i}/{m}', end='')
                print(f' - weight-norm: {np.linalg.norm(w)}')
        
        _, sig, VT = np.linalg.svd(W, full_matrices=0)
        indices = np.argsort(sig)[1:self.n_components + 1]
        Y = VT[indices, :] * np.sqrt(m)
        
        _, sig, VT = np.linalg.svd(Y, full_matrices=0)
        S = np.matrix(np.diag(sig ** (-1)))
        R = VT.T * S * VT
        self.Y, self.R = Y, R
        self._fitted = True
        return self
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return np.array(self.Y * self.R).T
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def _construct_graph(self, X: Matrix) -> Matrix:
        m, _ = X.shape
        kd_tree = KDTree(X.copy())
        index_mat = np.ones((m, self.n_neighbors), dtype=int)
        for i in range(m):
            _, neighbors = kd_tree.query(X[i], k=self.n_neighbors + 1, p=2)
            neighbors = neighbors.tolist()
            neighbors.remove(i)
            index_mat[i] = np.array([neighbors])
        
        return index_mat

    def set_params(self, 
                   n_neighbors: int = None, 
                   n_components: int = None,
                   regularization: float = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)
        if n_components is not None: self.n_components = int(n_components)
        if regularization is not None: self.regularization = float(regularization)


class SammonMapping(Transformer, Unsupervised):
    
    """
    Sammon mapping is a dimensionality reduction algorithm designed to transform 
    high-dimensional data into a lower-dimensional space. Sammon mapping focuses 
    on preserving the pairwise distances or dissimilarities between data points, 
    making it well-suited for capturing non-linear relationships in the data.
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    `max_iter` : Number of iteration
    `max_halves` : Number of halving of step-size
    `tol` : Threshold for early-stopping
    `initialize` : Methodology for embedding-space initialization
    
    """
    
    def __init__(self,
                 n_components: int = None,
                 max_iter: int = 100,
                 max_halves: int = 20,
                 tol: float = 1e-5,
                 initialize: Literal['pca'] | None = 'pca',
                 verbose: bool = False) -> None:
        self.n_components = n_components
        self.max_iter = max_iter
        self.max_halves = max_halves
        self.tol = tol
        self.initialize = initialize
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'SammonMapping':
        m, _ = X.shape
        X_dist = cdist(X, X)
        scale = 0.5 / X_dist.sum()
        
        Y = np.random.normal(0, 1, (m, self.n_components))
        if self.initialize == 'pca':
            Y = PCA(n_components=self.n_components).fit_transform(X)
            if self.verbose:
                print('[SammonMap] Finished PCA initialization')
        
        ones = np.ones((m, self.n_components))
        X_dist += np.eye(m)
        X_dist_inv = 1 / X_dist
        self._remove_inf(X_dist_inv, X_dist_inv)
        
        Y_dist = cdist(Y, Y) + np.eye(m)
        Y_dist_inv = 1 / Y_dist
        
        self._fill_diagnoal(X_dist, 1)
        self._fill_diagnoal(Y_dist, 1)
        self._fill_diagnoal(X_dist_inv, 0)
        self._fill_diagnoal(Y_dist_inv, 0)
        
        self._remove_inf(matrix=X_dist_inv, filter=X_dist_inv)
        self._remove_inf(matrix=Y_dist_inv, filter=X_dist_inv)
        self._remove_inf(matrix=X_dist_inv, filter=Y_dist_inv)
        self._remove_inf(matrix=Y_dist_inv, filter=Y_dist_inv)
        
        E = np.sum(((X_dist - Y_dist) ** 2) * X_dist_inv)
        for i in range(self.max_iter):
            delta = Y_dist_inv - X_dist_inv
            delta_one = delta.dot(ones)
            
            gradient = np.dot(delta, Y) - (Y * delta_one)
            hessian = np.dot(Y_dist_inv ** 3, Y ** 2) - delta_one
            hessian -= np.dot(2, Y) * np.dot(Y_dist_inv ** 3, Y)
            hessian += Y ** 2 * np.dot(Y_dist_inv ** 3, ones)
            
            step = -gradient.flatten(order='F') / np.abs(hessian.flatten(order='F'))
            Y_old = Y
            
            for j in range(self.max_halves):
                step_reshape = step.reshape(2, round(len(step) / 2)).T
                y = Y_old + step_reshape
                
                dist = cdist(Y, y) + np.eye(m)
                dist_inv = 1 / dist
                self._remove_inf(matrix=dist_inv, filter=dist_inv)
                
                delta = X_dist - dist
                E_new = np.sum((delta ** 2) * X_dist_inv)
                if E_new < E: break
                else: step = np.dot(0.5, step)
            
            if j == self.max_halves:
                print(f'[SammonMap] Warning: max halves reached.', end='')
                print(' Optimization may not converge!')
            
            if np.abs((E - E_new) / E) < self.tol:
                print(f'[SammonMap] Early convergence at iteration {i}/{self.max_iter}')
                break
            
            if self.verbose and i % 10 == 0:
                print(f'[SammonMap] Iteration: {i}/{self.max_iter}', end='')
                print(f' - relative delta-error: {np.abs((E - E_new) / E)}')
            
            E = E_new
        self.error = E * scale
        self.Y = y
        
        if self.verbose: print(f'[SammonMap] Final error: {self.error}')
        self._fitted = True
        return self
    
    def _fill_diagnoal(self, matrix: Matrix, value: float) -> None:
        np.fill_diagonal(matrix, value)
    
    def _remove_inf(self, matrix: Matrix, filter: Matrix) -> None:
        matrix[np.isinf(filter)] = 0
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.Y
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()

    def set_params(self,
                   n_components: int = None,
                   max_iter: int = None,
                   max_halves: int = None,
                   initialize: Literal = None,
                   tol: float = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if max_iter is not None: self.max_iter = int(max_iter)
        if max_halves is not None: self.max_halves = int(max_halves)
        if initialize is not None: self.initialize = str(initialize)
        if tol is not None: self.tol = float(tol)


class LaplacianEigenmap(Transformer, Unsupervised):
    
    """
    Laplacian Eigenmap is a dimensionality reduction technique that 
    transforms high-dimensional data into a lower-dimensional space while 
    preserving local relationships. It constructs a weighted graph based 
    on pairwise similarities between data points, computes the Laplacian 
    matrix from the graph, and performs eigenvalue decomposition to 
    obtain eigenvectors.
    
    Parameters
    ----------
    `n_components` : Dimensionality of the lower-dimensional space
    `sigma` : Width parameter for Gaussian kernel in affinity calculation
    
    """

    def __init__(self, n_components: int = None, sigma: float = 1.0) -> None:
        self.n_components = n_components
        self.sigma = sigma
        self.embedding = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'LaplacianEigenmap':
        pairwise = cdist(X, X)
        weights = np.exp(-pairwise / (2 * self.sigma ** 2))
        np.fill_diagonal(weights, 0)
        
        laplacian = self._compute_laplacian(weights)
        normalized_laplacian = self._normalize_laplacian(weights, laplacian)
        
        eigvals, eigvecs = np.linalg.eigh(normalized_laplacian)
        sorted_indices = np.argsort(eigvals)
        eigvals = eigvals[sorted_indices]
        eigvecs = eigvecs[:, sorted_indices]
        self.embedding = eigvecs[:, :self.n_components]
        self._fitted = True
        return self

    def _compute_laplacian(self, W: Matrix) -> Matrix:
        return np.diag(np.sum(W, axis=1)) - W
    
    def _normalize_laplacian(self, W: Matrix, L: Matrix) -> Matrix:
        D_sqrt_inv = np.linalg.inv(np.sqrt(np.diag(np.sum(W, axis=1))))
        return D_sqrt_inv.dot(L).dot(D_sqrt_inv)
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.embedding
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()

    def set_params(self, n_components: int = None, sigma: float = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if sigma is not None: self.sigma = float(sigma)


class Isomap(Transformer, Unsupervised):
    
    """
    Isomap, or Isometric Mapping, is a nonlinear dimensionality reduction technique 
    employed in machine learning. It focuses on preserving the geodesic distances 
    between data points in high-dimensional datasets. By constructing a neighborhood 
    graph and calculating pairwise geodesic distances, Isomap transforms the data 
    into a lower-dimensional space, offering insights into complex, nonlinear 
    relationships within the dataset.
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    `epsilon` : Boundary radius of the neighbor hypersphere
    `algorithm` : Shortest path finding algorithm
    (e.g. `dijkstra`, `floyd`)
    `metric` : Distance metric
    (e.g. `euclidean`, `cityblock`, `chebyshev`, `cosine`, `correlation`)
    
    """
    
    def __init__(self, 
                 n_components: int = None, 
                 epsilon: float = 5.0,
                 algorithm: Literal['dijkstra', 'floyd'] = 'dijkstra',
                 metric: Literal['euclidean', 'cityblock', 'chebyshev',
                                 'cosine', 'correlation'] = 'euclidean') -> None:
        self.n_components = n_components
        self.epsilon = epsilon
        self.algorithm = algorithm
        self.metric = metric
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'Isomap':
        self.adj = self._make_adjacency(X)
        if np.isinf(self.adj).any(): 
            print(f'[Isomap] Too narrow bin size with epsilon of {self.epsilon}!')
            return
        
        m, _ = self.adj.shape
        H = np.eye(m) - (1 / m) * np.ones((m, m))
        C = -0.5 / m * H.dot(self.adj ** 2).dot(H)
        
        eigvals, eigvecs = eigh(C)
        indices = eigvals.argsort()[::-1]
        eigvals, eigvecs = eigvals[indices], eigvecs[:, indices]
        self.eigvals = eigvals[:self.n_components]
        self.eigvecs = eigvecs[:, :self.n_components]
        self._fitted = True
        return self

    def _make_adjacency(self, X: Matrix) -> Matrix:
        m, _ = X.shape
        distance = cdist(X, X, metric=self.metric)
        adjacency = np.zeros((m, m)) + np.inf
        bin = distance < self.epsilon
        adjacency[bin] = distance[bin]
        
        if self.algorithm == 'dijkstra':
            path_mat = shortest_path(adjacency, directed=False, method='D')
        elif self.algorithm == 'floyd':
            path_mat = shortest_path(adjacency, directed=False, method='FW')
        else: raise UnsupportedParameterError(self.algorithm)
        return path_mat

    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.eigvecs.dot(np.diag(np.sqrt(self.eigvals)))
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def set_params(self, 
                   n_components: int = None, 
                   epsilon: float = None,
                   algorithm: Literal = None,
                   metric: Literal = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if epsilon is not None: self.epsilon = float(epsilon)
        if algorithm is not None: self.algorithm = str(algorithm)
        if metric is not None: self.method = str(metric)


class ConformalIsomap(Transformer, Unsupervised):
    
    """
    Conformal Isomap (C-Isomap) is a variation of the Isomap algorithm that 
    prioritizes preserving infinitesimal angles (conformality) by adjusting 
    geodesic distances to reflect scale information. Unlike traditional Isomap, 
    which focuses on preserving full isometry, C-Isomap is well-suited for 
    datasets with varying local scales or non-uniform structures.
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    `epsilon` : Boundary radius of the neighbor hypersphere
    `algorithm` : Shortest path finding algorithm
    (e.g. `dijkstra`, `floyd`)
    
    """
    
    def __init__(self,
                 n_components: int = None,
                 epsilon: float = 5.0,
                 algorithm: Literal['dijkstra', 'floyd'] = 'dijkstra') -> None:
        self.n_components = n_components
        self.epsilon = epsilon
        self.algorithm = algorithm
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'ConformalIsomap':
        self.adj = self._make_adjacency(X)
        if np.isinf(self.adj).any():
            print(f'[C-Isomap] Too narrow bin size with epsilon of {self.epsilon}!')
            return
        
        m, _ = self.adj.shape
        H = np.eye(m) - (1 / m) * np.ones((m, m))
        C = -0.5 / m * H.dot(self.adj ** 2).dot(H)
        
        eigvals, eigvecs = eigh(C)
        indices = eigvals.argsort()[::-1]
        eigvals, eigvecs = eigvals[indices], eigvecs[:, indices]
        self.eigvals = eigvals[:self.n_components]
        self.eigvecs = eigvecs[:, :self.n_components]
        self._fitted = True
        return self
    
    def _make_adjacency(self, X: Matrix) -> Matrix:
        m, _ = X.shape
        distance = cdist(X, X, metric='euclidean')
        adjacency = np.zeros((m, m)) + np.inf
        bin = distance < self.epsilon
        adjacency[bin] = distance[bin]
        
        masked = np.ma.masked_array(adjacency, mask=np.isinf(adjacency))
        masked_mean = np.ma.mean(masked, axis=0)
        for i in range(m):
            for j in range(m):
                adjacency[i, j] /= np.sqrt(masked_mean[i] * masked_mean[j])

        if self.algorithm == 'dijkstra':
            path_mat = shortest_path(adjacency, directed=False, method='D')
        elif self.algorithm == 'floyd':
            path_mat = shortest_path(adjacency, directed=False, method='FW')
        else: raise UnsupportedParameterError(self.algorithm)
        return path_mat
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.eigvecs.dot(np.diag(np.sqrt(self.eigvals)))
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()

    def set_params(self,
                   n_components: int = None,
                   epsilon: float = None,
                   algorithm: Literal = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if epsilon is not None: self.epsilon = float(epsilon)
        if algorithm is not None: self.algorithm = str(algorithm)


class LTSA(Transformer, Unsupervised):
    
    """
    Local Tangent Space Alignment (LTSA) is a dimensionality reduction technique 
    that aims to preserve the local geometry of data points in a manifold. 
    It works by approximating the tangent space at each data point and aligning 
    these local tangent spaces. This helps to capture the intrinsic structure 
    of the data, particularly in cases where the global geometry might be 
    complex or non-linear. 
    
    Parameters
    ----------
    `n_components` : Dimensionality of low-space
    `n_neighbors` : Number of neighbors
    
    """
    
    def __init__(self,
                 n_components: int = None,
                 n_neighbors: int = 10) -> None:
        self.n_components = n_components
        self.n_neighbors = n_neighbors
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'LTSA':
        m, n = X.shape
        self.neighbors, self.neighbor_indices = self._compute_neighborhood(X)
        M = np.zeros((m, m))
        use_svd = self.n_neighbors > n
        
        for i in range(m):
            Xi = self.neighbors[i]
            Xi -= np.mean(self.neighbors[i], axis=0)
            if use_svd:
                v = svd(Xi, full_matrices=True)[0]
            else:
                Ci = Xi.dot(Xi.T)
                v = eigh(Ci)[1][:, ::-1]
            
            Gi = np.zeros((self.n_neighbors, self.n_components + 1))
            Gi[:, 1:] = v[:, :self.n_components]
            Gi[:, 0] = 1 / np.sqrt(self.n_neighbors)
            
            neighbors_x, neighbors_y = np.meshgrid(self.neighbor_indices[i], 
                                                   self.neighbor_indices[i])
            M[neighbors_x, neighbors_y] -= Gi.dot(Gi.T)
            M[self.neighbor_indices[i], self.neighbor_indices[i]] += 1
        
        self._M = M
        self._fitted = True
        return self
    
    def _compute_neighborhood(self, X: Matrix) -> Matrix:
        distances = cdist(X, X, metric='euclidean')
        indices = np.argsort(distances, axis=1)[:, :self.n_neighbors]
        return X[indices], indices
    
    def _find_null_space(self, M: Matrix) -> Matrix:
        eigvals, eigvecs = eigh(M, subset_by_index=(1, self.n_components), 
                                overwrite_a=True)
        indices = np.argsort(np.abs(eigvals))
        return eigvecs[:, indices]
    
    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self._find_null_space(self._M)
    
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()

    def set_params(self,
                   n_components: int = None,
                   n_neighbors: int = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)




from abc import *
from typing import Any

from luma.core.main import LUMA


__all__ = (
    'Estimator', 
    'Transformer', 
    'Evaluator', 
    'Visualizer', 
    'Supervised', 
    'Unsupervised', 
    'Distance'
)


class Estimator(LUMA, metaclass=ABCMeta):
    
    """
    An estimator is a mathematical model or algorithm 
    used to make predictions or estimates based on data. 
    It takes input data and learns from it to generate output predictions, 
    typically for tasks like classification or regression. 
    Estimators can be as simple as linear regression or as complex as 
    deep neural networks, and their goal is to capture patterns and 
    relationships within the data to make accurate predictions on new, unseen data.
    
    Example
    -------
    >>> model = AnyEstimator(*args)
    >>> model.fit(X_train, y_train)
    >>> prediction = model.predict(X_test)
    
    """
    
    @abstractmethod
    def fit(self, *args) -> 'Estimator': ...
    
    @abstractmethod
    def predict(self, *args) -> ...: ...
    
    @abstractmethod
    def score(self, *args) -> float: ...
    
    @abstractmethod
    def set_params(self, *args) -> None: ...


class Transformer(LUMA, metaclass=ABCMeta):
    
    """
    A transformer (preprocessor) is a component or set of operations used to prepare and 
    clean raw data before it is fed into a machine learning model. 
    Preprocessing tasks can include data normalization, handling missing values, 
    feature scaling, one-hot encoding, and more. 
    The goal of a preprocessor is to make the data suitable for the specific 
    machine learning algorithm being used, enhancing the model's performance 
    by ensuring the data is in the right format and is free from inconsistencies or noise.
    
    Example
    -------
    >>> trans = AnyTransformer()
    >>> trans.fit(data)
    >>> new_data = trans.transform(data)
    
    To fit and transform at the same time,
    >>> new_data = trans.fit_transform(data)
    
    """
    
    class Feature: 
        
        """
        An inner class of `Transformer` dedicated to the processing and 
        handling of feature data in a dataset.
        
        Default class for uninherited models.
        """
        
    class Target:
        
        """
        An inner class of `Transformer` dedicated to the processing and 
        handling of target data in a dataset.
        """
         
    class Both: 
        
        """
        An inner class of `Transformer` dedicated to the processing and 
        handling of both feature and target data in a dataset.
        """
    
    @abstractmethod
    def fit(self, *args) -> 'Transformer': ...
    
    @abstractmethod
    def transform(self, *args) -> Any: ...
    
    @abstractmethod
    def fit_transform(self, *args) -> Any: ...
    
    @abstractmethod
    def set_params(self, *args) -> None: ...


class Evaluator(LUMA, metaclass=ABCMeta):
    
    """
    Evaluators, a.k.a. metrics are quantitative measures used to assess the performance 
    and effectiveness of a machine learning model. These metrics provide insights into 
    how well a model is performing on a particular task, 
    such as classification or regression. 
    
    Example
    -------
    >>> metric = AnyEvaluator()
    >>> score = metric.compute(y_true=target, y_pred=predictions)
    
    """
    
    @abstractstaticmethod
    def compute(*args) -> float: ...


class Visualizer(LUMA, metaclass=ABCMeta):
    
    """
    A visualizer is a tool that helps visualize and understand various aspects 
    of machine learning models, datasets, and the results the analysis. 
    Visualizers play a crucial role in simplifying the interpretation of complex 
    machine learning processes and results.
    
    Example
    -------
    >>> plotter = AnyVisualizer()
    >>> plotter.plot()
    
    """
    
    @abstractmethod
    def plot(self, *args) -> None: ...


class Supervised(LUMA):
    
    """
    Supervised learning is a type of machine learning where the algorithm learns 
    to make predictions or decisions by training on a labeled dataset. 
    In this approach, the algorithm is provided with input data and corresponding 
    target labels, and it learns to map the inputs to the correct outputs. 
    """
    
    def __init__(self, *args) -> None: ...


class Unsupervised(LUMA):
    
    """
    Unsupervised learning is a machine learning paradigm where the algorithm 
    is given input data without explicit target labels. Instead of making predictions, 
    the algorithm's goal is to discover hidden patterns, structures, 
    or relationships within the data.
    
    Default class for uninherited models.
    """
    
    def __init__(self, *args) -> None: ...
    
    @property
    def labels(self) -> Any:...


class Distance(LUMA, metaclass=ABCMeta):
    
    """
    In mathematics and machine learning, distance is a measure of how much "separation" 
    or "difference" there is between two points, objects, or distributions. Different 
    types of distances serve various purposes, and they are used in different contexts.
    
    Example
    -------
    >>> metric = AnyDistance()
    >>> dist = metric.distance(x, y)
    
    """
    
    @abstractstaticmethod
    def distance(*args) -> float: ...




from typing import *
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Estimator, Evaluator
from luma.interface.exception import NotFittedError
from luma.model_selection.cv import CrossValidator


__all__ = (
    'GridSearchCV'
)


class GridSearchCV:
    
    """
    Grid seach with cross validation(CV) is a method used to tune hyperparameters 
    of a model by searching through a specified parameter grid. It evaluates all 
    combinations of parameter values to find the best combination. This process 
    is performed using cross-validation to ensure model performance is not biased 
    to a specific data split. The result is the optimal set of parameters that 
    yield the best model performance according to a chosen metric.
    
    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `param_grid` : Parameter grid for repetitive search
    `cv` : K-fold size for cross validation
    `metric` : Scoring metric for evaluation
    `refit` : Whether to re-fit the estimator with the best parameters found
    `random_state` : Seed for random sampling for cross validation
    
    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * For `metric`, both class or instance are possible
    
    Examples
    --------
    >>> param_grid = {'param_1': [...], 
                      'param_2': [...],
                      ...,
                      AnyStr: List[Any]}
    
    >>> grid = GridSearchCV(estimator=AnyEstimator(),
                            param_grid=param_grid,
                            cv=5,
                            metric=AnyEvaluator,
                            refit=True,
                            random_state=None)
    >>> grid.fit(X, y)
    >>> score = grid.best_score
    >>> model = grid.best_model
    
    """
    
    def __init__(self, 
                 estimator: Estimator, 
                 param_grid: dict, 
                 cv: int = 5, 
                 metric: Evaluator = None,
                 refit: bool = True, 
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.param_grid = param_grid
        self.cv = cv
        self.metric = metric
        self.refit = refit
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> Estimator:
        best_score = None
        best_params = None
        param_combinations = self._get_param_combinations()
        max_iter = len(param_combinations)

        if self.verbose:
            print(f'Fitting {self.cv} folds for {max_iter} candidates,',
                  f'totalling {self.cv * max_iter} fits.\n')
        
        for i, params in enumerate(param_combinations, start=1):
            self.estimator.set_params(**params)
            cv_model = CrossValidator(estimator=self.estimator,
                                      metric=self.metric,
                                      cv=self.cv,
                                      random_state=self.random_state,
                                      verbose=self.verbose)
            
            mean_score = cv_model.score(X, y)
            if self.verbose:
                print(f'[GridSearchCV] candidate {i}/{max_iter} {params} - score:',
                      f'{mean_score:.3f}')
            if best_score is None or mean_score > best_score:
                best_score = mean_score
                best_params = params

        self.best_score = best_score
        self.best_params = best_params
        
        if self.verbose:
            print(f'\n[GridSearchCV] Best params: {self.best_params}')
            print(f'[GridSearchCV] Best score: {self.best_score}')
        
        if self.refit:
            self.estimator.set_params(**best_params)
            self.estimator.fit(X, y)
        
        self._fitted = True
        return self.best_model

    def _get_param_combinations(self) -> List[Dict[str, Any]]:
        keys, values = zip(*self.param_grid.items())
        param_combinations = np.array(np.meshgrid(*values)).T.reshape(-1, len(keys))
        param_combinations = [dict(zip(keys, v)) for v in param_combinations]
        return param_combinations
    
    @property
    def best_model(self) -> Estimator:
        if not self._fitted: raise NotFittedError(self)
        return self.estimator




from typing import Tuple
from collections import Counter
import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError
from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import TreeNode
from luma.metric.classification import Accuracy


__all__ = (
    'DecisionTreeClassifier'
)


class DecisionTreeClassifier(Estimator, Supervised):
    
    """
    A Decision Tree Classifier is a supervised machine learning algorithm 
    used for classification tasks. It works by recursively partitioning 
    the input data into subsets based on the values of different features. 
    At each node of the tree, a decision is made based on a specific feature, 
    and the data is split into branches accordingly.
    
    Parameters
    ----------
    `max_depth` : Maximum depth of tree
    `min_samples_split` : Minimum number of samples required to split a node
    
    """

    def __init__(self, 
                 max_depth: int = 100,
                 min_samples_split: int = 2,
                 verbose: bool = False) -> None:
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.verbose = verbose
        self.root = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'DecisionTreeClassifier':
        _, self.n_features = X.shape
        self.root = self._grow_tree(X, y)
        
        self._fitted = True
        return self

    def _grow_tree(self, X: Matrix, y: Matrix, 
                   depth: int = 0) -> TreeNode:
        _, n = X.shape
        if self._stopping_criteria(X, y, depth):
            leaf_value = self._most_common_label(y)
            return TreeNode(value=leaf_value)

        feature_indices = np.random.choice(n, self.n_features, replace=False)
        
        best_feature, best_thresh = self._best_criteria(X, y, feature_indices)
        left_indices, right_indices = self._split(X[:, best_feature], best_thresh)
        
        if len(left_indices) == 0 or len(right_indices) == 0:
            leaf_value = self._most_common_label(y)
            return TreeNode(value=leaf_value)
        
        left = self._grow_tree(X[left_indices], y[left_indices], depth + 1)
        right = self._grow_tree(X[right_indices], y[right_indices], depth + 1)
        
        if self.verbose:
            print(f'[DecisionTree] Depth {depth} reached -', end=' ')
            print(f'best_feature: {best_feature}, best_threshold: {best_thresh}')
        
        return TreeNode(best_feature, best_thresh, left, right)
    
    def _stopping_criteria(self, X: Matrix, y: Matrix, 
                           depth: int) -> bool:
        if depth >= self.max_depth: return True
        if len(np.unique(y)) == 1: return True
        if X.shape[0] < self.min_samples_split: return True
        return False

    def _best_criteria(self, X: Matrix, y: Matrix, 
                       indices: Matrix) -> Tuple[int, float]:
        best_gain = -1
        split_idx, split_thresh = 0, 0.0
        for idx in indices:
            X_col = X[:, idx]
            thresholds = np.unique(X_col)
            for threshold in thresholds:
                gain = self._information_gain(X_col, y, threshold)

                if gain > best_gain:
                    best_gain = gain
                    split_idx = idx
                    split_thresh = threshold

        return split_idx, split_thresh

    def _information_gain(self, X_col: Matrix, y: Matrix, 
                          thresh: float) -> float:
        parent_entropy = self._entropy(y)
        left_indices, right_indices = self._split(X_col, thresh)
        if not len(left_indices) or not len(right_indices): return 0

        n = len(y)
        n_l, n_r = len(left_indices), len(right_indices)
        e_l, e_r = self._entropy(y[left_indices]), self._entropy(y[right_indices])
        child_entropy = (n_l / n) * e_l + (n_r / n) * e_r

        ig = parent_entropy - child_entropy
        return ig

    def _split(self, X_col: Matrix, thresh: float) -> Tuple[Matrix]:
        left_indices = np.argwhere(X_col <= thresh).flatten()
        right_indices = np.argwhere(X_col > thresh).flatten()
        return left_indices, right_indices

    def _traverse_tree(self, x: Matrix, node: TreeNode) -> float | None:
        if node.isLeaf: return node.value
        if x[node.feature] <= node.threshold:
            return self._traverse_tree(x, node.left)
        return self._traverse_tree(x, node.right)

    def _most_common_label(self, y: Matrix) -> int:
        counter = Counter(y)
        most_common = counter.most_common(1)[0][0]
        return most_common
    
    def _entropy(self, y: Matrix) -> float:
        hist = np.bincount(y)
        ps = hist / len(y)
        return -np.sum([p * np.log2(p) for p in ps if p])
    
    def print_tree(self) -> None:
        if not self._fitted: raise NotFittedError(self)
        self._print_tree_recursive(self.root, depth=0)
    
    def _print_tree_recursive(self, node: TreeNode, depth: int) -> None:
        if node.isLeaf: print(f"{(' ' * 4 + '|') * depth}--- Leaf: {node.value}")
        else:
            print(f"{(' ' * 4 + '|') * depth}---", 
                  f"Feature {node.feature} <= {node.threshold}")
            self._print_tree_recursive(node.left, depth + 1)
            self._print_tree_recursive(node.right, depth + 1)

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return np.array([self._traverse_tree(x, self.root) for x in X])

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.compute(y_true=y, y_pred=X_pred)

    def set_params(self,
                   max_depth: int = None,
                   min_samples_split: int = None) -> None:
        if max_depth is not None: self.max_depth = int(max_depth)
        if min_samples_split is not None: 
            self.min_samples_split = int(min_samples_split)




from typing import Any
from luma.interface.super import Transformer, Estimator


__all__ = (
    'NotFittedError', 
    'UnsupportedParameterError', 
    'NotConvergedError',
    'ModelExtensionError'
)


def get_name(model: Any) -> str: return type(model).__name__


class NotFittedError(Exception):
    def __init__(self, model: Transformer | Estimator | Any) -> None:
        super().__init__(f"'{get_name(model)}' is not fitted!" + 
                         f" Call '{get_name(model)}.fit()' to fit the model.")
        

class UnsupportedParameterError(Exception):
    def __init__(self, param: Any) -> None:
        super().__init__(f"'{param}' is unsupported!")


class NotConvergedError(Exception):
    def __init__(self, model: Transformer | Estimator | Any) -> None:
        super().__init__(f"'{get_name(model)}' did not converged!" + 
                         f" Try setting '{get_name(model)}.tol' to bigger value.")


class ModelExtensionError(Exception):
    def __init__(self, filename: str) -> None:
        super().__init__(f"'{filename}' is not a '.luma' model file!")




import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Evaluator


__all__ = (
    'Accuracy', 
    'Precision', 
    'Recall', 
    'F1Score', 
    'Specificity', 
    'AUCCurveROC', 
    'Complex'
)


class Accuracy(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(y_true == y_pred)


class Precision(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        true_positives = np.sum((y_true == 1) & (y_pred == 1))
        false_positives = np.sum((y_true == 0) & (y_pred == 1))
        return true_positives / (true_positives + false_positives)


class Recall(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        true_positives = np.sum((y_true == 1) & (y_pred == 1))
        false_negatives = np.sum((y_true == 1) & (y_pred == 0))
        return true_positives / (true_positives + false_negatives)


class F1Score(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        precision = Precision.compute(y_true, y_pred)
        recall = Recall.compute(y_true, y_pred)
        return 2 * (precision * recall) / (precision + recall)


class Specificity(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        true_negatives = np.sum((y_true == 0) & (y_pred == 0))
        false_positives = np.sum((y_true == 0) & (y_pred == 1))
        return true_negatives / (true_negatives + false_positives)


class AUCCurveROC(Evaluator):
    @staticmethod
    def roc_curve(y_true: Matrix, y_scores: Matrix) -> float:
        thresholds = np.sort(y_scores)
        tpr, fpr = [], []
        for threshold in thresholds:
            y_pred = y_scores >= threshold
            tpr.append(Recall.compute(y_true, y_pred))
            fpr.append(1 - Specificity.compute(y_true, y_pred))
        return fpr, tpr
    
    @staticmethod
    def auc_roc(fpr: float, tpr: float) -> float:
        n = len(fpr)
        auc = 0
        for i in range(1, n):
            auc += (fpr[i] - fpr[i - 1]) * tpr[i]
        return auc


class Complex:
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> dict:
        scores = dict()
        scores['accuracy'] = Accuracy.compute(y_true, y_pred)
        scores['precision'] = Precision.compute(y_true, y_pred)
        scores['recall'] = Recall.compute(y_true, y_pred)
        scores['f1-score'] = F1Score.compute(y_true, y_pred)
        scores['specificity'] = Specificity.compute(y_true, y_pred)
        return scores




from scipy.spatial.distance import pdist, squareform
import matplotlib.pyplot as plt
import numpy as np

from luma.interface.super import Evaluator
from luma.interface.util import Matrix, SilhouetteUtil, DBUtil


__all__ = (
    'SilhouetteCoefficient', 
    'DaviesBouldin'
)


class SilhouetteCoefficient(Evaluator):
    
    """
    The Silhouette Coefficient is a measure used to evaluate the quality of 
    clusters in a dataset. It calculates how similar each data point is to its 
    own cluster compared to other clusters. The coefficient ranges from -1 
    (poorly clustered) to +1 (well clustered), with values near 0 indicating 
    overlapping clusters. High average scores across a dataset suggest clear, 
    well-separated clusters.
    
    Parameters
    ----------
    `data` : Original data
    `labels` : Labels assigned by clustering estimator
    
    Examples
    --------
    With Instantiation
    >>> sil = SilhouetteCoefficient(data, labels)
    >>> score = sil.compute(data, labels) # compute() is a static method
    >>> sil.plot()
    
    Without Instantiation
    >>> score = SilhouetteCoefficient.compute(data, labels)
    >>> SilhouetteCoefficient.plot() # Error; plot() is an instance method
    
    """
    
    def __init__(self, data: Matrix, labels: Matrix) -> None:
        self.data = data
        self.labels = labels
        self.distances = squareform(pdist(self.data))
    
    @staticmethod
    def compute(data: Matrix, labels: Matrix) -> Matrix[float]:
        scores = []
        distances = squareform(pdist(data))
        for idx, label in enumerate(labels):
            util = SilhouetteUtil(idx, label, labels, distances)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            scores.append(score)

        return np.mean(scores)
    
    def _individual_silhouette(self) -> Matrix:
        silhouette_values = []
        for idx, label in enumerate(self.labels):
            util = SilhouetteUtil(idx, label, self.labels, self.distances)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            silhouette_values.append(score)
        
        return np.array(silhouette_values)

    def plot(self) -> None:
        sample_silhouette = self._individual_silhouette()
        y_lower = 10
        
        for i in range(len(set(self.labels))):
            values = sample_silhouette[self.labels == i]
            values.sort()

            size_cluster_i = values.shape[0]
            y_upper = y_lower + size_cluster_i

            color = plt.cm.nipy_spectral(float(i) / len(set(self.labels)))
            plt.fill_betweenx(np.arange(y_lower, y_upper), 0, values,
                              facecolor=color, edgecolor=color, alpha=0.7)

            y_lower = y_upper + 10
        
        plt.title("Silhouette Coefficient Plot")
        plt.xlabel("Silhouette coefficient values")
        plt.ylabel("Cluster label")
        plt.axvline(x=self.compute(self.data, self.labels), 
                    color="red", linestyle="--")
        
        plt.yticks([])
        plt.xticks(np.arange(0.0, 1.2, 0.1))
        plt.tight_layout()
        plt.show()


class DaviesBouldin(Evaluator):
    
    """
    The Davies-Bouldin Index (DBI) is a metric for evaluating clustering 
    algorithms. It compares the average distance within clusters to the 
    distance between clusters. Lower DBI values indicate better clustering, 
    with compact and well-separated clusters.
    
    Parameters
    ----------
    `data` : Original data
    `labels` : Labels assigned by clustering estimator
    
    """
    
    @staticmethod
    def compute(data: Matrix, labels: Matrix) -> float:
        util = DBUtil(data=data, labels=labels)
        centroids = util.cluster_centroids
        scatter = util.within_cluster_scatter
        separation = util.separation

        n_clusters = len(centroids)
        db_values = np.zeros(n_clusters)

        for i in range(n_clusters):
            max_ratio = 0
            for j in range(n_clusters):
                if i == j: continue
                ratio = (scatter[i] + scatter[j]) / separation[i, j]
                if ratio > max_ratio: max_ratio = ratio
            
            db_values[i] = max_ratio
        db_index = np.mean(db_values)
        
        return db_index




from typing import Dict, Literal, Tuple, List, TypeVar, Union, Any

from luma.interface.super import Estimator, Transformer, Evaluator
from luma.interface.super import Supervised
from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError

T = TypeVar('T', bound=Union[Estimator, Transformer])


__all__ = (
    'Pipeline'
)


class Pipeline:
    
    """
    A pipeline is a sequence of steps to process data and build a model. 
    It includes data collection and preprocessing, model selection and training, 
    evaluation, and deployment. This systematic approach ensures efficiency 
    and accuracy in model development. Pipelines are often automated for 
    scalability and reproducibility in real-world applications.
    
    Parameters
    ----------
    `models` : List of models \n
    You can assign labels to each model by encapsulating the label and the model
    inside a tuple. \n
    Otherwise, the name of the model is automatically assigned by default. \n
    
    `param_dict` : Dictionary of parameters for each models \n
    You must specify the name(or label) of the model and its parameter name
    in the key of the dictionary. \n
    e.g. `{'model_name__param_name': value}`
    
    Examples
    --------
    >>> pipe = Pipeline(
            models=[
                ('trans_1', Transformer()),
                ('trans_2', Transformer()),
                ...,
                ('est', Estimator())
            ],
            param_dict={
                'trans_1__param': Any,
                'trans_2__param': Any,
                ...,
                'est__param': Any
            }
        )
    >>> pipe.fit(X_train, y_train)
    >>> y_pred = pipe.predict(X_test)
    
    Notes
    -----
    * To use `Pipeline` with visual methods of `luma.visual`, make sure to
    transform data using `pipe.transform()` if the pipeline sequence contains
    transformers.
    
    * Not all the models are compatible with `Pipeline`.

    """
    
    def __init__(self,
                 models: List[Tuple[str, T]] | List[T],
                 param_dict: Dict[str, Any] = dict(),
                 verbose: bool = False) -> None:
        self.models: Dict[T] = dict()
        self.param_dict = param_dict
        self.verbose = verbose
        self._X: Matrix
        self._y: Matrix
        self._fitted = False

        for model in models:
            if isinstance(model, tuple):
                _name, _model = model
                self.models[_name] = _model
            else: self.models[type(model).__name__] = model
        
        self.set_params(param_dict)
    
    def fit(self, X: Matrix, y: Matrix) -> 'Pipeline':
        self._X, self._y = self.fit_transform(X, y)
        model = self.estimator
        if hasattr(model, 'verbose'):
            model.verbose = self.verbose
        
        data = [self._X]
        if isinstance(model, Supervised): data.append(self._y)
        model.fit(*data)
        
        self._fitted = True
        return self
        
    def transform(self, X: Matrix, y: Matrix) -> Tuple[Matrix, Matrix]:
        if not self._fitted: raise NotFittedError(self)
        X_trans, y_trans = X, y
        for model in self.transformers:
            if isinstance(model, Transformer.Target):
                y_trans = model.transform(y_trans)
            elif isinstance(model, Transformer.Both):
                X_trans, y_trans = model.transform(X_trans)
            else:
                X_trans = model.transform(X_trans)

        return X_trans, y_trans
    
    def predict(self, X: Matrix, transform: bool = True):
        if not self._fitted: raise NotFittedError(self)
        for model in self.transformers:
            if not transform: break
            if isinstance(model, Transformer.Target | Transformer.Both):
                continue
            
            X = model.transform(X)
        
        return self.estimator.predict(X)
    
    def fit_transform(self, X: Matrix, y: Matrix) -> Tuple[Matrix, Matrix]:
        X_trans, y_trans = X, y
        for model in self.transformers:
            data = [X_trans]
            if hasattr(model, 'verbose'):
                model.verbose = self.verbose
            
            if isinstance(model, Supervised):
                data.append(y_trans)
            if isinstance(model, Transformer.Target):
                y_trans = model.fit_transform(y_trans)
            elif isinstance(model, Transformer.Both):
                X_trans, y_trans = model.fit_transform(*data)
            else:
                X_trans = model.fit_transform(*data)

        return X_trans, y_trans
    
    def fit_predict(self, X: Matrix, y: Matrix) -> Matrix:
        self.fit(X, y)
        return self.predict(X)
    
    def set_params(self, param_dict: Dict[str, Any]) -> None:
        for param_name, value in param_dict.items():
            try: _name, _param = param_name.split('__')
            except: raise UnsupportedParameterError(param_name)
            self.models[_name].set_params(**{_param: value})
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator | Literal['default'] = 'default') -> float:
        model = self.estimator
        pkg = model.__module__.split('.')[1]
        if metric == 'default':
            if pkg == 'classifier': metric = Accuracy
            elif pkg == 'regressor': metric = MeanSquaredError
        
        X_pred = self.predict(X)
        return metric.compute(y_true=y, y_pred=X_pred)
    
    def dump(self) -> None:
        print(f"Configuration of a pipeline:")
        for name, model in self.models.items():
            print(f"\n[{type(model).__name__} as '{name}']")
            for attr in model.__dict__.items():
                print(*attr, sep=': ')
    
    @property
    def transformers(self) -> List[Transformer]:
        _trans = []
        for model in self.models.values():
            if isinstance(model, Transformer):
                _trans.append(model)

        return _trans

    @property
    def estimator(self) -> Estimator:
        for model in self.models.values():
            if isinstance(model, Estimator):
                return model
    
    @property
    def transformed_data(self) -> Tuple[Matrix, Matrix]:
        return self._X, self._y
    
    def __getitem__(self, index: int) -> T:
        for i, model in enumerate(self.models.values()):
            if index == i: return model
        else: raise IndexError('Model index out of bounds!')
    
    def __setitem__(self, label: str, model: T):
        self.models[label] = model




import numpy as np

from luma.interface.super import Estimator, Evaluator, Unsupervised
from luma.interface.util import Matrix, Vector
from luma.metric.clustering import SilhouetteCoefficient
from luma.interface.exception import NotFittedError


__all__ = (
    'GaussianMixture',
    'MultinomialMixture'
)


class GaussianMixture(Estimator, Unsupervised):
    
    """
    A Gaussian Mixture Model (GMM) is a probabilistic model for representing 
    an ensemble of multiple Gaussian distributions within a dataset. It is 
    used in clustering by assuming each cluster follows a different Gaussian 
    distribution. GMM provides soft clustering, assigning probabilities to 
    each data point for belonging to each cluster. The model parameters are 
    estimated using the Expectation-Maximization algorithm.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `max_iter` : Maximum amount of iteration
    `tol` : Tolerance for early convergence
    `random_state` : Seed for random permutation
    
    """
    
    def __init__(self, 
                 n_clusters : int,
                 max_iter: int = 100,
                 tol: float = 1e-5,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.verbose = verbose
        self._X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'GaussianMixture':
        self._X = X
        self._initialize_params(X)
        for i in range(self.max_iter):
            R = self._E_step(X)
            means_old = self.means.copy()
            
            self._M_step(X, R)
            norm_ = np.linalg.norm(self.means - means_old)
            
            if self.verbose and i % 20 and i:
                print(f'[GMM] Finished iteration {i}/{self.max_iter}', end='')
                print(f' with delta-means-norm: {norm_}')
            
            if norm_ < self.tol:
                if self.verbose:
                    print(f'[GMM] Early-convergence at iteration', end='')
                    print(f' {i}/{self.max_iter}')
                break
        
        self._fitted = True
        return self
    
    def _initialize_params(self, X: Matrix) -> None:
        m, _ = X.shape
        np.random.seed(self.random_state)
        random_indices = np.random.permutation(m)[:self.n_clusters]
        
        self.means = X[random_indices]
        self.covs = np.array([np.cov(X.T) for _ in range(self.n_clusters)])
        self.weights = np.full(self.n_clusters, 1 / self.n_clusters)
    
    def _E_step(self, X: Matrix) -> Matrix:
        m, n = X.shape
        resp = np.zeros((m, self.n_clusters))
        
        for i in range(self.n_clusters):
            cov_inv = np.linalg.inv(self.covs[i])
            diff = X - self.means[i]
            exp_term = np.exp(-0.5 * np.sum(diff.dot(cov_inv) * diff, axis=1))
            coeff = 1 / np.sqrt(np.linalg.det(self.covs[i]) * (2 * np.pi) ** n)
            resp[:, i] = self.weights[i] * coeff * exp_term
        
        resp /= resp.sum(axis=1, keepdims=True)
        return resp
    
    def _M_step(self, X: Matrix, R: Matrix) -> None:
        self.weights = R.mean(axis=0)
        self.means = R.T.dot(X) / np.sum(R, axis=0)[:, np.newaxis]
        for i in range(self.n_clusters):
            diff = X - self.means[i]
            self.covs[i] = (R[:, i, np.newaxis] * diff).T.dot(diff) / R[:, i].sum()
    
    @property
    def labels(self) -> Vector:
        return self.predict(self._X)
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        R = self._E_step(X)
        return np.argmax(R, axis=1)
    
    def predict_proba(self, X: Matrix) -> Vector:
        return self._E_step(X)
    
    def score(self, X: Matrix, metric: Evaluator = SilhouetteCoefficient) -> float:
        X_pred = self.predict(X)
        return metric.compute(X, X_pred)
    
    def set_params(self, 
                   n_clusters: int = None,
                   max_iter: int = None,
                   tol: float = None,
                   random_state: int = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if max_iter is not None: self.max_iter = int(max_iter)
        if tol is not None: self.tol = float(tol)
        if random_state is not None: self.random_state = int(random_state)


class MultinomialMixture(Estimator, Unsupervised):
    
    """
    Multinomial Mixture Models (MMM) cluster categorical data using a mixture 
    of multinomial distributions. Each cluster is modeled with a distinct 
    multinomial distribution, representing the probability of each category 
    within that cluster. The Expectation-Maximization (EM) algorithm is 
    typically used to estimate model parameters, including cluster 
    probabilities and multinomial parameters. MMM is commonly applied in text 
    analysis and other scenarios involving discrete data, like document 
    clustering or topic modeling.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `max_iter` : Maximum amount of iteration
    `tol` : Tolerance for early convergence
    
    """
    
    def __init__(self, 
                 n_clusters: int, 
                 max_iter: int = 100, 
                 tol: float = 1e-5,
                 verbose: bool = False):
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self._X = None
        self._fitted = False
        
    def fit(self, X: Matrix) -> 'MultinomialMixture':
        self._X = X
        m, _ = X.shape
        
        prev_logL = None
        self._initialize_parameters(X)
        for i in range(self.max_iter):
            resp = self._E_step(X)
            self._M_step(X, resp)
            
            logL = 0
            for i in range(m):
                likelihoods = []
                for j in range(self.n_clusters):
                    cluster_likelihood = np.prod(self.theta[j] ** X[i, :])
                    weighted_likelihood = self.pi[j] * cluster_likelihood
                    likelihoods.append(weighted_likelihood)

                L_sum = np.sum(likelihoods)
                logL += np.log(L_sum)
            
            if prev_logL is not None:
                diff = np.abs(logL - prev_logL)
                if self.verbose and i % 20 == 0 and i:
                    print(f'[MMM] Finished iteration {i}/{self.max_iter}', end='')
                    print(f' with delta-likelihood: {diff}')
                    
                if diff < self.tol:
                    if self.verbose:
                        print(f'[MMM] Early-convergence at {i}/{self.max_iter}', end='')
                        print(f' with delta-likelihood: {diff}')
                    break
            
            prev_logL = logL
        
        self._fitted = True
        return self
    
    def _initialize_parameters(self, X: Matrix) -> None:
        _, n = X.shape
        self.pi = np.ones(self.n_clusters) / self.n_clusters
        self.theta = np.random.dirichlet(alpha=np.ones(n), size=self.n_clusters)
    
    def _E_step(self, X: Matrix) -> Matrix:
        n_samples = X.shape[0]
        resp = np.zeros((n_samples, self.n_clusters))
        
        for j in range(self.n_clusters):
            likelihood = np.prod(self.theta[j] ** X, axis=1)
            resp[:, j] = self.pi[j] * likelihood
        
        resp /= resp.sum(axis=1, keepdims=True)
        return resp
    
    def _M_step(self, X: Matrix, R: Matrix) -> None:
        self.pi = R.mean(axis=0)
        for j in range(self.n_clusters):
            weighted_sum = np.dot(R[:, j], X)
            self.theta[j] = weighted_sum / np.sum(R[:, j])

    @property
    def labels(self) -> Vector:
        return self.predict(self._X)
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        R = self._E_step(X)
        return np.argmax(R, axis=1)
    
    def score(self, X: Matrix, metric: Evaluator = SilhouetteCoefficient) -> float:
        X_pred = self.predict(X)
        return metric.compute(X, X_pred)
    
    def set_params(self, 
                   n_clusters: int = None,
                   max_iter: int = None,
                   tol: float = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if max_iter is not None: self.max_iter = int(max_iter)
        if tol is not None: self.tol = float(tol)




from typing import List, Literal
import numpy as np

from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector, Scalar, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.preprocessing.encoder import LabelEncoder


__all__ = (
    'MajorityVoteClassifier'
)


class MajorityVoteClassifier(Estimator, Supervised):
    
    """
    A majority vote classifier is an ensemble method that combines predictions 
    from multiple machine learning models. It aggregates the predictions from 
    each model and selects the final output based on the majority vote. The 
    voting can be based on either the predicted class labels or the predicted 
    probabilities. This approach often leads to improved model performance and 
    robustness compared to individual models.
    
    Parameters
    ----------
    `estimators` : List of estimators to vote
    `voting` : Voting critetion
    (`label` to vote the most frequent and `prob` to vote the most probable)
    `weights` : Weights for each classifier on voting
    (`None` for uniform weights)
    
    Examples
    --------
    >>> maj = MajorityVoteClassifier(estimators=[AnyEstimator(), ...],
                                     voting='label',
                                     weights=[0.25, 0.5, ...])
    >>> maj.fit(X, y)
    >>> pred = maj.predict(X)
    >>> est = maj[i] # Get i-th estimator from `maj`
    
    """
    
    def __init__(self,
                 estimators: List[Estimator] = None,
                 voting: Literal['label', 'prob'] = 'label',
                 weights: Vector[Scalar] = None,
                 verbose: bool = True) -> None:
        self.estimators = estimators
        self.voting = voting
        self.weights = weights
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'MajorityVoteClassifier':
        if self.voting not in ('label', 'prob'):
            raise UnsupportedParameterError(self.voting)
        
        if self.weights and len(self.weights) != len(self.estimators):
            raise ValueError(f"Size mismatch between 'weights' and 'estimators'!")
        
        self.le = LabelEncoder()
        self.le.fit(y)
        
        self.classes = self.le.classes
        self.estimators_ = []
        
        for est in self.estimators:
            fitted_est = Clone(est).get
            if hasattr(fitted_est, 'verbose'):
                fitted_est.verbose = self.verbose
            
            fitted_est.fit(X, self.le.transform(y))
            self.estimators_.append(fitted_est)
        
        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        
        if self.voting == 'label':
            preds = np.asarray([est.predict(X) for est in self.estimators_]).T
            majority_vote = np.apply_along_axis(
                lambda x: np.argmax(np.bincount(x, weights=self.weights)),
                axis=1,
                arr=preds
            )
        else:
            majority_vote = np.argmax(self.predict_proba(X), axis=1)
        
        majority_vote = self.le.inverse_transform(majority_vote)
        return majority_vote
    
    def predict_proba(self, X: Matrix) -> Matrix:
        probas = []
        for est in self.estimators_:
            if not hasattr(est, 'predict_proba'):
                raise ValueError(f"'{type(est)}' does not support 'predict_proba'")
            probas.append(est.predict_proba(X))
        
        return np.average(probas, axis=0, weights=self.weights)

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.compute(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   estimators: List[Estimator] = None,
                   voting: Literal = None,
                   weights: Vector[Scalar] = None) -> None:
        if estimators is not None: self.estimators = estimators
        if voting is not None: self.voting = str(voting)
        if weights is not None: self.weights = weights

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]




from typing import *
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Evaluator


__all__ = (
    'MeanAbsoluteError', 
    'MeanSquaredError', 
    'RootMeanSquaredError',
    'MeanAbsolutePercentageError', 
    'RSquaredScore', 
    'Complex'
)


class MeanAbsoluteError(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(np.abs(y_true - y_pred))


class MeanSquaredError(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean((y_true - y_pred) ** 2)


class RootMeanSquaredError(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        return np.sqrt(np.mean((y_true - y_pred) ** 2))


class MeanAbsolutePercentageError(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


class RSquaredScore(Evaluator):
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> float:
        y_bar = np.mean(y_true)
        total_sum_of_squares = np.sum((y_true - y_bar) ** 2)
        residual_sum_of_squares = np.sum((y_true - y_pred) ** 2)
        r2 = 1 - (residual_sum_of_squares / total_sum_of_squares)
        return r2


class Complex:
    @staticmethod
    def compute(y_true: Matrix, y_pred: Matrix) -> dict:
        scores = dict()
        scores['mae'] = MeanAbsoluteError.compute(y_true, y_pred)
        scores['mse'] = MeanSquaredError.compute(y_true, y_pred)
        scores['rmse'] = RootMeanSquaredError.compute(y_true, y_pred)
        scores['mape'] = MeanAbsolutePercentageError.compute(y_true, y_pred)
        scores['rsqe'] = RSquaredScore.compute(y_true, y_pred)
        return scores




from typing import Literal, Tuple
from scipy.spatial.distance import pdist, squareform
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Estimator, Evaluator, Unsupervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.clustering.kmeans import KMeansClusteringPlus
from luma.metric.clustering import SilhouetteCoefficient


__all__ = (
    'AgglomerativeClustering', 
    'DivisiveClustering'
)


class AgglomerativeClustering(Estimator, Unsupervised):
    
    """
    Agglomerative clustering is a hierarchical clustering technique that starts 
    by treating each data point as a single cluster. It iteratively merges the 
    closest pairs of clusters based on a chosen distance metric. This process 
    continues until a specified number of clusters is reached or all points 
    are merged into a single cluster.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `linkage` : Linkage method (e.g. `single`, `complete`, `average`, `ward`)
    
    Examples
    --------
    >>> agg = AgglomerativeClustering()
    >>> agg.fit(X, y)
    >>> lables = agg.labels # Get assigned labels
    
    """
    
    def __init__(self, 
                 n_clusters: int = 2, 
                 linkage: Literal['single', 'complete', 
                                  'average', 'ward'] = 'single'):
        self.n_clusters = n_clusters
        self.linkage = linkage
        self._X = None
        self._fitted = False

    def fit(self, X: Matrix) -> 'AgglomerativeClustering':
        self._X = X
        n_samples = X.shape[0]
        dist_matrix = squareform(pdist(X, metric='euclidean'))
        clusters = {i: [i] for i in range(n_samples)}
        
        while len(clusters) > self.n_clusters:
            closest_pair, _ = self._find_closest_clusters(dist_matrix, clusters)

            self._merge_clusters(clusters, closest_pair)
            self._update_distance_matrix(dist_matrix, closest_pair, clusters)

        self.clusters = clusters
        self._n = n_samples
        
        self._fitted = True
        return self

    def _find_closest_clusters(self, 
                               dist_matrix: Matrix, 
                               clusters: dict) -> Tuple[tuple, float]:
        min_dist = np.inf
        closest_pair = (None, None)

        for i in clusters:
            for j in clusters:
                if i != j and dist_matrix[i, j] < min_dist:
                    min_dist = dist_matrix[i, j]
                    closest_pair = (i, j)

        return closest_pair, min_dist

    def _merge_clusters(self, clusters: Matrix, pair: tuple) -> None:
        i, j = pair
        clusters[i].extend(clusters[j])
        del clusters[j]

    def _update_distance_matrix(self, 
                                dist_matrix: Matrix, 
                                pair: tuple, 
                                clusters: dict) -> None:
        i, j = pair
        for k in clusters:
            if k == i: continue
            value = None
            if self.linkage == 'single':
                value = min(dist_matrix[i, k], dist_matrix[j, k])
            elif self.linkage == 'complete':
                value = max(dist_matrix[i, k], dist_matrix[j, k])
            elif self.linkage == 'average':
                value = np.mean([dist_matrix[i, k], dist_matrix[j, k]])
            else:
                raise UnsupportedParameterError(self.linkage)
            
            dist_matrix[i, k] = dist_matrix[k, i] = value
            
        dist_matrix[i, i] = np.inf

    def _assign_labels(self, clusters: dict, n_samples: int) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        
        labels = np.zeros(n_samples, dtype=int)
        for cluster_id, cluster_samples in enumerate(clusters.values()):
            for sample in cluster_samples:
                labels[sample] = cluster_id
        
        return labels
    
    @property
    def labels(self) -> Matrix:
        return self._assign_labels(self.clusters, self._n)

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")
    
    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.compute(self._X, self.labels)
    
    def set_params(self, 
                   n_clusters: int = None,
                   linkage: Literal = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if linkage is not None: self.linkage = str(linkage)


class DivisiveClustering(Estimator, Unsupervised):
    
    """
    Divisive clustering is a "top-down" approach in hierarchical clustering, 
    starting with all data points in one large cluster. It iteratively splits 
    clusters into smaller ones based on similarity measures, typically using 
    algorithms like K-means for splitting. The process continues until a 
    specified number of clusters is reached or other stopping criteria are met. 
    This method is particularly effective for datasets with well-defined, 
    separate clusters.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    
    Examples
    --------
    >>> div = DivisiveClustering()
    >>> div.fit(X)
    >>> labels = div.labels # Get assigned labels
    
    """
    
    def __init__(self, 
                 n_clusters: int = 2):
        self.n_clusters = n_clusters
        self.clusters = []
        self._X = None
        self._fitted = False

    def fit(self, X: Matrix) -> 'DivisiveClustering':
        initial_cluster = [0] * X.shape[0]
        self._X = X
        self.clusters.append(initial_cluster)
        self._recursive_split(X)
        
        self._fitted = True
        return self

    def _recursive_split(self, X: Matrix, depth: int = 0) -> None:
        if depth == self.n_clusters - 1: return
        
        max_cluster_idx = self._find_largest_cluster()
        split_idx = [i for i, v in enumerate(self.clusters[-1]) 
                            if v == max_cluster_idx]

        if len(split_idx) > 1:
            kmeans = KMeansClusteringPlus(n_clusters=2).fit(X[split_idx])
            new_labels = kmeans.predict(X[split_idx])

            new_cluster = self.clusters[-1].copy()
            new_cluster_id = max(new_cluster) + 1
            for idx, label in zip(split_idx, new_labels):
                new_cluster[idx] = new_cluster_id \
                    if label == 1 else max_cluster_idx

            self.clusters.append(new_cluster)
            self._recursive_split(X, depth + 1)

    def _find_largest_cluster(self) -> Matrix:
        return np.argmax(np.bincount(self.clusters[-1]))

    @property
    def labels(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return np.array(self.clusters[-1])
    
    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")
    
    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.compute(self._X, self.labels)
    
    def set_params(self, n_clusters: int = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)




import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="luma-ml",
    version='0.5.1',
    author="ChanLumerico",
    author_email="greensox284@gmail.com",
    description="A Comprehensive Python Module for Machine Learning and Data Science",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChanLumerico/LUMA",
    packages=setuptools.find_namespace_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'matplotlib',
        'seaborn'
    ]
)

"""
Terminal
--------
>>> python setup.py sdist bdist_wheel
>>> python -m twine upload dist/*

"""



from typing import Literal, Optional
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap 
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Visualizer, Estimator


__all__ = (
    'DecisionRegion', 
    'ClusterPlot'
)


class DecisionRegion(Visualizer):
    def __init__(self, 
                 estimator: Estimator,
                 X: Matrix, 
                 y: Optional[Matrix] = None, 
                 title: str | Literal['auto'] = 'auto', 
                 xlabel: str = r'$x_1$', 
                 ylabel: str = r'$x_2$',
                 cmap: ListedColormap = 'rainbow',
                 alpha: float = 0.4) -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha
        
        if self.title == 'auto':
            self.title = type(self.estimator).__name__

        if self.y is None and hasattr(self.estimator, 'labels'):
            self.y = self.estimator.labels

    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             size: float = 250, 
             scale: float = 10.0, 
             show: bool = False) -> plt.Axes:
        x1_min, x1_max = self.X[:, 0].min(), self.X[:, 0].max()
        x2_min, x2_max = self.X[:, 1].min(), self.X[:, 1].max()
        delta_1, delta_2 = (x1_max - x1_min) / size, (x2_max - x2_min) / size
        
        x1_min, x1_max = x1_min - delta_1 * scale, x1_max + delta_1 * scale
        x2_min, x2_max = x2_min - delta_2 * scale, x2_max + delta_2 * scale
        xx1, xx2 = np.meshgrid(np.arange(x1_min, x1_max, delta_1), 
                               np.arange(x2_min, x2_max, delta_2))
        
        Z = self.estimator.predict(np.array([xx1.ravel(), xx2.ravel()]).T)
        Z = Z.reshape(xx1.shape)
        
        if ax is None: _, ax = plt.subplots()
        ax.contourf(xx1, xx2, Z, 
                    alpha=self.alpha, 
                    cmap=self.cmap, 
                    levels=len(np.unique(self.y)))
        
        ax.set_xlim(xx1.min(), xx1.max())
        ax.set_ylim(xx2.min(), xx2.max())

        ax.scatter(self.X[:, 0], self.X[:, 1], 
                   c=self.y, 
                   cmap=self.cmap, 
                   alpha=0.8, 
                   edgecolors='black')
        
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_title(self.title)
        ax.figure.tight_layout()
        
        if show: plt.show()
        return ax


class ClusterPlot(Visualizer):
    def __init__(self,
                 estimator: Estimator,
                 X: Matrix,
                 title: str | Literal['auto'] = 'auto',
                 xlabel: str = r'$x_1$',
                 ylabel: str = r'$x_2$',
                 cmap: ListedColormap = 'rainbow',
                 alpha: float = 0.8) -> None:
        self.estimator = estimator
        self.X = X
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha
        self.labels = self.estimator.labels
        
        if self.title == 'auto':
            self.title = type(self.estimator).__name__
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        ax.scatter(self.X[self.labels == -1, 0], 
                   self.X[self.labels == -1, 1],
                   marker='x',
                   c='black', 
                   label='Noise')
        
        ax.scatter(self.X[self.labels != -1, 0], 
                   self.X[self.labels != -1, 1], 
                   marker='o',
                   c=self.labels[self.labels != -1],
                   cmap=self.cmap,
                   alpha=self.alpha,
                   edgecolors='black')
        
        ax.set_title(self.title)
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        
        if len(self.X[self.labels == -1]): ax.legend()
        ax.figure.tight_layout()

        if show: plt.show()
        return ax




from typing import *
from scipy.stats import mode
import numpy as np

from luma.interface.util import Matrix
from luma.classifier.tree import DecisionTreeClassifier
from luma.regressor.tree import DecisionTreeRegressor
from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError


__all__ = (
    'RandomForestClassifier', 
    'RandomForestRegressor'
)


class RandomForestClassifier(Estimator, Supervised):
    
    """
    A Random Forest Classifier is an ensemble learning method in machine learning, 
    widely used for both classification and regression tasks. It operates by 
    constructing a multitude of decision trees during the training phase and 
    outputs the mode of the classes for classification.
    
    Parameters
    ----------
    `n_trees` : Number of trees in the forest
    `max_depth` :  Maximum depth of each trees
    `min_samples_split` : Minimum number of samples required to split a node
    `bootstrap` : Whether to bootstrap the samples of dataset
    `bootstrap_feature` : Whether to bootstrap the features of each data
    `n_features` : Number of features to be sampled when bootstrapping features
    
    """
    
    def __init__(self, 
                 n_trees: int = 10, 
                 max_depth: int = 100,
                 min_samples_split: int = 2,
                 bootstrap: bool = True,
                 bootstrap_feature: bool = False,
                 n_features: int | Literal['auto'] = 'auto',
                 verbose: bool = False) -> None:
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_features = n_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.verbose = verbose
        self.trees = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'RandomForestClassifier':
        m, n = X.shape
        if self.n_features == 'auto': 
            self.n_features = int(n ** 0.5)
        else:
            if isinstance(self.n_features, str):
                raise UnsupportedParameterError(self.n_features)
        
        self.trees = [DecisionTreeClassifier() for _ in range(self.n_trees)]
        for i, tree in enumerate(self.trees, start=1):
            X_sample, y_sample = X, y
            if self.bootstrap:
                bootstrap_data = np.random.choice(m, m, replace=True)
                X_sample = X_sample[bootstrap_data]
                y_sample = y_sample[bootstrap_data]
            if self.bootstrap_feature:
                bootstrap_feature = np.random.choice(n, self.n_features, replace=True)
                X_sample = X_sample[:, bootstrap_feature]
            
            tree.set_params(max_depth=self.max_depth, 
                            min_samples_split=self.min_samples_split)
            tree.fit(X_sample, y_sample)
            
            if self.verbose:
                print(f'[RandomForest] Finished fitting tree {i}/{self.n_trees}')
        
        self._fitted = True
        return self
    
    def print_forest(self) -> None:
        if not self._fitted: raise NotFittedError(self)
        for i, tree in enumerate(self.trees, start=1):
            print(f'[Tree {i}/{self.n_trees}]')
            tree.print_tree()

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        preds = np.array([tree.predict(X) for tree in self.trees])
        majority, _ = mode(preds, axis=0)
        
        return majority.flatten()
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.compute(y_true=y, y_pred=X_pred)
    
    def set_params(self,
                   n_trees: int = None,
                   max_depth: int = None,
                   min_samples_split: int = None,
                   n_features: int | str = None,
                   bootstrap: bool = None,
                   bootstrap_feature: bool = None) -> None:
        if n_trees is not None: self.n_trees = int(n_trees)
        if max_depth is not None: self.max_depth = int(max_depth)
        if bootstrap is not None: self.bootstrap = bool(bootstrap)
        
        if bootstrap_feature is not None: 
            self.bootstrap_feature = bool(bootstrap_feature)
        if min_samples_split is not None: 
            self.min_samples_split = int(min_samples_split)
        if n_features is not None:
            if isinstance(n_features, int):
                self.n_features = int(n_features)
            elif isinstance(n_features, str):
                self.n_features = str(n_features)


class RandomForestRegressor(Estimator, Supervised):
    
    """
    A Random Forest Regressor is an ensemble learning algorithm for regression 
    tasks. It builds multiple decision trees during training, each on a random 
    subset of features and bootstrapped data. The final prediction is the average 
    of individual tree predictions, providing a robust and accurate result.
    
    Parameters
    ----------
    `n_trees` : Number of trees in the forest
    `max_depth` :  Maximum depth of each trees
    `min_samples_split` : Minimum number of samples required to split a node
    `bootstrap` : Whether to bootstrap the samples of dataset
    `bootstrap_feature` : Whether to bootstrap the features of each data
    `n_features` : Number of features to be sampled when bootstrapping features
    
    """
    
    def __init__(self, 
                 n_trees: int = 10, 
                 max_depth: int = 100,
                 min_samples_split: int = 2,
                 bootstrap: bool = True,
                 bootstrap_feature: bool = False,
                 n_features: int | Literal['auto'] = 'auto',
                 verbose: bool = False) -> None:
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_features = n_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.verbose = verbose
        self.trees = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'RandomForestRegressor':
        m, n = X.shape
        if self.n_features == 'auto': 
            self.n_features = int(n ** 0.5)
        else:
            if isinstance(self.n_features, str):
                raise UnsupportedParameterError(self.n_features)
        
        self.trees = [DecisionTreeRegressor() for _ in range(self.n_trees)]
        for i, tree in enumerate(self.trees, start=1):
            X_sample, y_sample = X, y
            if self.bootstrap:
                bootstrap_data = np.random.choice(m, m, replace=True)
                X_sample = X_sample[bootstrap_data]
                y_sample = y_sample[bootstrap_data]
            if self.bootstrap_feature:
                bootstrap_feature = np.random.choice(n, self.n_features, replace=True)
                X_sample = X_sample[:, bootstrap_feature]
            
            tree.set_params(max_depth=self.max_depth, 
                            min_samples_split=self.min_samples_split)
            tree.fit(X_sample, y_sample)
            
            if self.verbose:
                print(f'[RandomForest] Finished fitting tree {i}/{self.n_trees}')
        
        self._fitted = True
        return self
    
    def print_forest(self) -> None:
        if not self._fitted: raise NotFittedError(self)
        for i, tree in enumerate(self.trees, start=1):
            print(f'[Tree {i}/{self.n_trees}]')
            tree.print_tree()

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        preds = np.array([tree.predict(X) for tree in self.trees])
        average_predictions = np.mean(preds, axis=0)
        
        return average_predictions
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.compute(y_true=y, y_pred=X_pred)
    
    def set_params(self,
                   n_trees: int = None,
                   max_depth: int = None,
                   min_samples_split: int = None,
                   n_features: int | str = None,
                   bootstrap: bool = None,
                   bootstrap_feature: bool = None) -> None:
        if n_trees is not None: self.n_trees = int(n_trees)
        if max_depth is not None: self.max_depth = int(max_depth)
        if bootstrap is not None: self.bootstrap = bool(bootstrap)
        
        if bootstrap_feature is not None: 
            self.bootstrap_feature = bool(bootstrap_feature)
        if min_samples_split is not None: 
            self.min_samples_split = int(min_samples_split)
        if n_features is not None:
            if isinstance(n_features, int):
                self.n_features = int(n_features)
            elif isinstance(n_features, str):
                self.n_features = str(n_features)




import numpy as np

from luma.interface.super import Estimator, Evaluator
from luma.interface.util import Matrix, Vector


__all__ = (
    'CrossValidator'
)


class CrossValidator(Evaluator):
    def __init__(self,
                 estimator: Estimator,
                 metric: Evaluator,
                 cv: int = 5,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.metric = metric
        self.cv = cv
        self.random_state = random_state
        self.verbose = verbose
    
    def score(self, X: Matrix, y: Vector) -> float:
        num_samples = X.shape[0]
        fold_size = num_samples // self.cv
        scores = []

        indices = np.arange(num_samples)
        np.random.shuffle(indices)
        
        for i in range(self.cv):
            start = i * fold_size
            end = start + fold_size if i < self.cv - 1 else num_samples

            test_indices = indices[start:end]
            train_indices = np.concatenate((indices[0:start], indices[end:]))

            X_train, X_test = X[train_indices], X[test_indices]
            y_train, y_test = y[train_indices], y[test_indices]

            self.estimator.fit(X_train, y_train)
            if self.metric: score = self.estimator.score(X_test, y_test, metric=self.metric)
            else: score = self.estimator.score(X_test, y_test)
            scores.append(score)
        
            if self.verbose:
                print(f'[CV] fold {i + 1} - score: {score:.3f}')
        
        return np.mean(scores)

    def compute(self, X: Matrix, y: Vector) -> float:
        return self.score(X, y)

